/**
 * rijndael-cbc.c from rijndael-api-fst.c
 *
 * @version 2.9 (December 2000)
 *
 * Optimised ANSI C code for the Rijndael cipher (now AES)
 *
 * @author Vincent Rijmen <vincent.rijmen@esat.kuleuven.ac.be>
 * @author Antoon Bosselaers <antoon.bosselaers@esat.kuleuven.ac.be>
 * @author Paulo Barreto <paulo.barreto@terra.com.br>
 *
 * This code is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Acknowledgements:
 *
 * We are deeply indebted to the following people for their bug reports,
 * fixes, and improvement suggestions to this implementation. Though we
 * tried to list all contributions, we apologise in advance for any
 * missing reference.
 *
 * Andrew Bales <Andrew.Bales@Honeywell.com>
 * Markus Friedl <markus.friedl@informatik.uni-erlangen.de>
 * John Skodon <skodonj@webquill.com>
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tinyaes/tinyaes.h"

int
aes_cbc_crypt(aes_ctx *ctx, u8 IV[MAX_IV_SIZE],
	      u8 *outBuffer, u8 *input, size_t inputLen, int mode)
{
	int i, numBlocks, padLen;
	u8 block[16], *iv;

	if (input == NULL || inputLen <= 0) {
		return 0; /* nothing to do */
	}

	numBlocks = inputLen / 16;
	
	iv = IV;
	if (mode) {
		for (i = numBlocks; i > 0; i--) {
			((u32*)block)[0] = ((u32*)input)[0] ^ ((u32*)iv)[0];
			((u32*)block)[1] = ((u32*)input)[1] ^ ((u32*)iv)[1];
			((u32*)block)[2] = ((u32*)input)[2] ^ ((u32*)iv)[2];
			((u32*)block)[3] = ((u32*)input)[3] ^ ((u32*)iv)[3];
			rijndaelEncrypt(ctx->ek, ctx->Nr, block, outBuffer);
			memcpy(iv, outBuffer, 16);
			input += 16;
			outBuffer += 16;
		}
	} else {
		if (inputLen % 16 != 0) {
			return -1;
		}

		for (i = numBlocks; i > 0; i--) {
			rijndaelDecrypt(ctx->dk, ctx->Nr, input, block);
			((u32*)block)[0] ^= ((u32*)iv)[0];
			((u32*)block)[1] ^= ((u32*)iv)[1];
			((u32*)block)[2] ^= ((u32*)iv)[2];
			((u32*)block)[3] ^= ((u32*)iv)[3];
			memcpy(iv, input, 16);
			memcpy(outBuffer, block, 16);
			input += 16;
			outBuffer += 16;
		}
	}
	return 16 * numBlocks;
}

int
aes_cbc_crypt_finish(aes_ctx *ctx, u8 IV[MAX_IV_SIZE],
			  u8 *outBuffer, u8 *input, size_t inputLen, int mode)
{
	int i, numBlocks, padLen;
	u8 block[16], *iv;
	u8 *p;

	if (input == NULL || inputLen <= 0) {
		return 0; /* nothing to do */
	}

	numBlocks = inputLen / 16;

	iv = IV;
	if (mode) {
		for (i = numBlocks; i > 0; i--) {
			((u32*)block)[0] = ((u32*)input)[0] ^ ((u32*)iv)[0];
			((u32*)block)[1] = ((u32*)input)[1] ^ ((u32*)iv)[1];
			((u32*)block)[2] = ((u32*)input)[2] ^ ((u32*)iv)[2];
			((u32*)block)[3] = ((u32*)input)[3] ^ ((u32*)iv)[3];
			rijndaelEncrypt(ctx->ek, ctx->Nr, block, outBuffer);
			iv = outBuffer;
			input += 16;
			outBuffer += 16;
		}
		padLen = 16 - (inputLen - 16 * numBlocks);
		if (!(padLen > 0 && padLen <= 16))
			return -1;
		for (i = 0; i < 16 - padLen; i++) {
			block[i] = input[i] ^ iv[i];
		}
		for (i = 16 - padLen; i < 16; i++) {
			block[i] = (u8)padLen ^ iv[i];
		}
		rijndaelEncrypt(ctx->ek, ctx->Nr, block, outBuffer);
		return 16 * (numBlocks + 1);
	} else {
		/* decrypt */
		if (inputLen % 16 != 0) {
			return -1;
		}

		/* all blocks but last */
		for (i = numBlocks - 1; i > 0; i--) {
			rijndaelDecrypt(ctx->dk, ctx->Nr, input, block);
			((u32*)block)[0] ^= ((u32*)iv)[0];
			((u32*)block)[1] ^= ((u32*)iv)[1];
			((u32*)block)[2] ^= ((u32*)iv)[2];
			((u32*)block)[3] ^= ((u32*)iv)[3];
			memcpy(iv, input, 16);
			memcpy(outBuffer, block, 16);
			input += 16;
			outBuffer += 16;
		}
		/* last block */
		rijndaelDecrypt(ctx->dk, ctx->Nr, input, block);
		((u32*)block)[0] ^= ((u32*)iv)[0];
		((u32*)block)[1] ^= ((u32*)iv)[1];
		((u32*)block)[2] ^= ((u32*)iv)[2];
		((u32*)block)[3] ^= ((u32*)iv)[3];
		padLen = block[15];
		if (padLen <= 0 || padLen > 16) {
			return -1;
		}
		for (i = 16 - padLen; i < 16; i++) {
			if (block[i] != padLen) {
				return -1;
			}
		}
		memcpy(outBuffer, block, 16 - padLen);
		return 16 * numBlocks - padLen;
	}
}
